/*
 * SPDX-License-Identifier:     GPL-2.0+
 *
 * Copyright (c) 2023 Unfolded Circle ApS
 */

#include <common.h>
#include <dm.h>
#include <linux/delay.h>
#include <asm/gpio.h>
#include <asm/io.h>
#include <power/fuel_gauge.h>
#include <power/pmic.h>

extern void uc_show_bmp(const char *filename);

DECLARE_GLOBAL_DATA_PTR;

#define LED_TOP_ROW 111
#define LED_BOTTOM_ROW 112
#define MAGNET_SENSOR 6

#define BUTTON_UP 20
#define BUTTON_DOWN 23
#define BUTTON_OUT 99

#define MIN_VOLTAGE_LEVEL 3550

#define SPI_SDI 102
#define SPI_SCL 110
#define SPI_CS 103
#define LCD_RST 115
#define ICN_EN 147

enum boot_images
{
	UCLOGO,
	LOWBAT,
	OPENERROR,
	RECOVERY
};

void keyboard_backlight(bool enable)
{
	gpio_request(LED_TOP_ROW, "Led top row");
	gpio_direction_output(LED_TOP_ROW, enable);
	gpio_free(LED_TOP_ROW);

	gpio_request(LED_BOTTOM_ROW, "Led bottom row");
	gpio_direction_output(LED_BOTTOM_ROW, enable);
	gpio_free(LED_BOTTOM_ROW);
}

void flash_led(int times)
{
	int i = 0;

	while (i < times)
	{
		keyboard_backlight(false);
		mdelay(500);
		keyboard_backlight(true);
		mdelay(500);
		i++;
	}
}

int check_if_device_open(void)
{
	gpio_request(MAGNET_SENSOR, "Magnet sensor");
	gpio_direction_input(MAGNET_SENSOR);

	int value = gpio_get_value(MAGNET_SENSOR);
	printf("Magnet sensor is: %d\n", value);

	gpio_free(MAGNET_SENSOR);

	return value;
}

int check_battery(void)
{
	struct udevice *fg;
	int ret;

	ret = uclass_first_device_err(UCLASS_FG, &fg);
	if (ret)
	{
		if (ret == -ENODEV)
			printf("Can't find fuel gauge\n");
		else
			printf("Get UCLASS_FG failed: %d\n", ret);
		return ret;
	}

	int voltage = fuel_gauge_get_voltage(fg);
	printf("Battery voltage: %d\n", voltage);

	int soc = fuel_gauge_update_get_soc(fg);
	printf("SOC: %d\n", soc);

	return voltage < MIN_VOLTAGE_LEVEL ? 0 : 1;
}

int check_charging(void)
{
	struct udevice *fg;
	int ret;

	ret = uclass_first_device_err(UCLASS_FG, &fg);
	if (ret)
	{
		if (ret == -ENODEV)
			printf("Can't find fuel gauge\n");
		else
			printf("Get UCLASS_FG failed: %d\n", ret);
		return ret;
	}

	// give some time to the IC to catch up
	mdelay(500);

	int current = fuel_gauge_get_current(fg);
	printf("Current: %d\n", current);

	return current > 0 ? 1 : 0;
}

int check_recovery(void)
{
	gpio_request(BUTTON_OUT, "Button out");
	gpio_direction_output(BUTTON_OUT, 1);

	gpio_request(BUTTON_UP, "Button up");
	gpio_direction_input(BUTTON_UP);

	int value_up = gpio_get_value(BUTTON_UP);
	printf("Button up is: %d\n", value_up);

	gpio_free(BUTTON_UP);

	gpio_request(BUTTON_DOWN, "Button down");
	gpio_direction_input(BUTTON_DOWN);

	int value_down = gpio_get_value(BUTTON_DOWN);
	printf("Button down is: %d\n", value_down);

	gpio_free(BUTTON_DOWN);

	gpio_free(BUTTON_OUT);

	return value_up && value_down ? 1 : 0;
}

int shutdown_pmic(void)
{
	struct udevice *pmic;
	int ret;

	printf("Shutting down PMIC\n");

	ret = uclass_first_device_err(UCLASS_PMIC, &pmic);

	if (ret)
	{
		if (ret == -ENODEV)
			printf("Can't find PMIC\n");
		else
			printf("Get UCLASS_PMIC failed: %d\n", ret);
		return ret;
	}
	else
	{
		return pmic_shutdown(pmic);
	}
}

void init_icn6211(void)
{
	// reset the bridge chip
	gpio_request(ICN_EN, "icn_enable");
	gpio_direction_output(ICN_EN, 1);
	gpio_free(ICN_EN);

	mdelay(100);

	// send init commands via i2c
	run_command("i2c dev 5", 0);
	run_command("i2c mw 0x2c 0x20, 0x56", 0);
	run_command("i2c mw 0x2c 0x21, 0xE0", 0);
	run_command("i2c mw 0x2c 0x22, 0x13", 0);
	run_command("i2c mw 0x2c 0x23, 0x19", 0);
	run_command("i2c mw 0x2c 0x24, 0x02", 0);
	run_command("i2c mw 0x2c 0x25, 0x7A", 0);
	run_command("i2c mw 0x2c 0x26, 0x00", 0);
	run_command("i2c mw 0x2c 0x27, 0x16", 0);
	run_command("i2c mw 0x2c 0x28, 0x02", 0);
	run_command("i2c mw 0x2c 0x29, 0x26", 0);
	run_command("i2c mw 0x2c 0x34, 0x80", 0);
	run_command("i2c mw 0x2c 0x36, 0x19", 0);
	run_command("i2c mw 0x2c 0x86, 0x29", 0);
	run_command("i2c mw 0x2c 0xB5, 0xA0", 0);
	run_command("i2c mw 0x2c 0x5C, 0xFF", 0);
	run_command("i2c mw 0x2c 0x2A, 0x07", 0);
	run_command("i2c mw 0x2c 0x56, 0x92", 0);
	run_command("i2c mw 0x2c 0x6B, 0x72", 0);
	run_command("i2c mw 0x2c 0x69, 0x27", 0);
	run_command("i2c mw 0x2c 0x10, 0x50", 0);
	run_command("i2c mw 0x2c 0x11, 0x98", 0);
	run_command("i2c mw 0x2c 0xB6, 0x20", 0);
	run_command("i2c mw 0x2c 0x51, 0x20", 0);
	run_command("i2c mw 0x2c 0x09, 0x10", 0);
}

void ams326pm01_spi_write_byte(u8 cmd)
{
	u32 kk;

	gpio_set_value(SPI_SCL, 0);

	for (kk = 0; kk < 8; kk++)
	{
		if (cmd & 0x80)
		{
			gpio_set_value(SPI_SDI, 1);
		}
		else
		{
			gpio_set_value(SPI_SDI, 0);
		}
		cmd <<= 1;
		gpio_set_value(SPI_SCL, 0);
		ndelay(50);
		gpio_set_value(SPI_SCL, 1);
		ndelay(50);
	}
};

void ams326pm01_spi_write_cmd(u8 addr, u8 value)
{
	gpio_set_value(SPI_CS, 0);
	ams326pm01_spi_write_byte(addr);
	ams326pm01_spi_write_byte(value);
	gpio_set_value(SPI_CS, 1);
};

void display_init(void)
{
	gpio_request(LCD_RST, "lcd_reset");
	gpio_direction_output(LCD_RST, 1);

	gpio_request(SPI_CS, "spi_cs");
	gpio_direction_output(SPI_CS, 1);

	gpio_request(SPI_SCL, "spi_scl");
	gpio_direction_output(SPI_SCL, 0);

	gpio_request(SPI_SDI, "spi_sdi");
	gpio_direction_output(SPI_SDI, 0);

	// reset
	gpio_set_value(LCD_RST, 0);
	mdelay(10);
	gpio_set_value(LCD_RST, 1);
	mdelay(100);

	/* panel condition set */
	ams326pm01_spi_write_cmd(0x27, 0x02);
	ams326pm01_spi_write_cmd(0x28, 0x88);
	ams326pm01_spi_write_cmd(0x29, 0x65);
	ams326pm01_spi_write_cmd(0x2a, 0x90);
	ams326pm01_spi_write_cmd(0x2b, 0x55);
	ams326pm01_spi_write_cmd(0x2c, 0x21);
	ams326pm01_spi_write_cmd(0x2d, 0x31);
	ams326pm01_spi_write_cmd(0x2e, 0x10);
	ams326pm01_spi_write_cmd(0x2f, 0x41);
	ams326pm01_spi_write_cmd(0x30, 0x21);
	ams326pm01_spi_write_cmd(0x36, 0x05);
	ams326pm01_spi_write_cmd(0x37, 0x07);
	ams326pm01_spi_write_cmd(0x38, 0x1c);
	ams326pm01_spi_write_cmd(0x39, 0x1c);
	ams326pm01_spi_write_cmd(0x3a, 0x1c);
	ams326pm01_spi_write_cmd(0x3b, 0x1c);

	/* display condition set */
	ams326pm01_spi_write_cmd(0x05, 0x0f);
	ams326pm01_spi_write_cmd(0x15, 0x00);
	ams326pm01_spi_write_cmd(0x19, 0x03);
	ams326pm01_spi_write_cmd(0x1a, 0x56);
	ams326pm01_spi_write_cmd(0x1d, 0x26);
	ams326pm01_spi_write_cmd(0x1f, 0x7a);
	ams326pm01_spi_write_cmd(0x3c, 0x57);
	ams326pm01_spi_write_cmd(0x3e, 0x16);
	ams326pm01_spi_write_cmd(0x3f, 0x0c);
	ams326pm01_spi_write_cmd(0x40, 0x06);
	ams326pm01_spi_write_cmd(0x41, 0x06);
	ams326pm01_spi_write_cmd(0x42, 0x06);

	/* pentile gamma condition set */
	ams326pm01_spi_write_cmd(0xff, 0x01);
	ams326pm01_spi_write_cmd(0x0a, 0x12);
	ams326pm01_spi_write_cmd(0x0b, 0x12);
	ams326pm01_spi_write_cmd(0x0c, 0x1c);
	ams326pm01_spi_write_cmd(0x0d, 0x1c);
	ams326pm01_spi_write_cmd(0x0e, 0x1b);
	ams326pm01_spi_write_cmd(0x0f, 0x13);
	ams326pm01_spi_write_cmd(0x10, 0x24);
	ams326pm01_spi_write_cmd(0x11, 0x21);
	ams326pm01_spi_write_cmd(0x2b, 0x12);
	ams326pm01_spi_write_cmd(0x2c, 0x12);
	ams326pm01_spi_write_cmd(0x2d, 0x1c);
	ams326pm01_spi_write_cmd(0x2e, 0x1c);
	ams326pm01_spi_write_cmd(0x2f, 0x1b);
	ams326pm01_spi_write_cmd(0x30, 0x13);
	ams326pm01_spi_write_cmd(0x31, 0x24);
	ams326pm01_spi_write_cmd(0x32, 0x21);
	ams326pm01_spi_write_cmd(0x4c, 0x12);
	ams326pm01_spi_write_cmd(0x4d, 0x12);
	ams326pm01_spi_write_cmd(0x4e, 0x1c);
	ams326pm01_spi_write_cmd(0x4f, 0x1c);
	ams326pm01_spi_write_cmd(0x50, 0x1b);
	ams326pm01_spi_write_cmd(0x51, 0x13);
	ams326pm01_spi_write_cmd(0x52, 0x24);
	ams326pm01_spi_write_cmd(0x53, 0x21);
	ams326pm01_spi_write_cmd(0xff, 0x00);

	ndelay(100);

	// gamma ctrl
	ams326pm01_spi_write_cmd(0x93, 0x01);
	ams326pm01_spi_write_cmd(0x60, 0x00);
	ams326pm01_spi_write_cmd(0x61, 0xa8);
	ams326pm01_spi_write_cmd(0x62, 0xc4);
	ams326pm01_spi_write_cmd(0x63, 0xdb);
	ams326pm01_spi_write_cmd(0x64, 0xbc);
	ams326pm01_spi_write_cmd(0x65, 0xc7);
	ams326pm01_spi_write_cmd(0x66, 0x6f);
	ams326pm01_spi_write_cmd(0x67, 0x00);
	ams326pm01_spi_write_cmd(0x68, 0x9a);
	ams326pm01_spi_write_cmd(0x69, 0xc1);
	ams326pm01_spi_write_cmd(0x6a, 0xdb);
	ams326pm01_spi_write_cmd(0x6b, 0xb9);
	ams326pm01_spi_write_cmd(0x6c, 0xc7);
	ams326pm01_spi_write_cmd(0x6d, 0x6e);
	ams326pm01_spi_write_cmd(0x6f, 0x8c);
	ams326pm01_spi_write_cmd(0x70, 0xdb);
	ams326pm01_spi_write_cmd(0x71, 0xd8);
	ams326pm01_spi_write_cmd(0x72, 0xb2);
	ams326pm01_spi_write_cmd(0x73, 0xc3);
	ams326pm01_spi_write_cmd(0x74, 0xa7);
	ams326pm01_spi_write_cmd(0x93, 0x00);

	mdelay(100);

	// sleep out
	ams326pm01_spi_write_cmd(0x00, 0xA0);
	mdelay(250);

	gpio_free(LCD_RST);
	gpio_free(SPI_CS);
	gpio_free(SPI_SCL);
	gpio_free(SPI_SDI);
}

void show_image(enum boot_images image)
{
	switch (image)
	{
	case UCLOGO:
		uc_show_bmp("img/boot_logo.bmp");
		break;

	case LOWBAT:
		uc_show_bmp("img/low_battery.bmp");
		break;

	case OPENERROR:
		uc_show_bmp("img/open_error.bmp");
		break;

	case RECOVERY:
		uc_show_bmp("img/recovery.bmp");
		break;

	default:
		uc_show_bmp("img/boot_logo.bmp");
		break;
	}

	init_icn6211();
	display_init();
}

int board_early_init_r(void)
{
	return 0;
}

int misc_init_r(void)
{
	keyboard_backlight(true);

	// if battery is low and not charging, stop boot
	if (!check_battery() && !check_charging())
	{
		printf("Low battery, aborting boot\n");

		// show low battery screen
		show_image(LOWBAT);

		// flash LEDs 5 times
		flash_led(5);

		// then turn off pmic
		return shutdown_pmic();
	}

	// check if we need to enter recovery mode
	if (check_recovery())
	{
		printf("Recovery boot selected\n");
		show_image(RECOVERY);
		run_command("run recoverybootcmd", 0);
	}

	show_image(UCLOGO);

	return 0;
}
